@@ -5,7 +5,7 @@
 ;; Author: rahguzar <rahguzar@zohomail.eu>
 ;; Maintainer: rahguzar <rahguzar@zohomail.eu>
 ;; Created: May 20, 2023
-;; Version: 0.1.2
+;; Version: 0.2.0
 ;; Keywords: convenience files tools unix
 ;; Homepage: https://codeberg.org/rahguzar/filechooser
 ;; Package-Requires: ((emacs "28.1") (compat "29.1"))
@@ -15,9 +15,7 @@
 ;;; Commentary:
 ;; An implementation of xdg-desktop-portal filechooser in Emacs. This allows
 ;; for choosing files in applications like firefox (with GTK_USE_PORTAL set)
-;; from an Emacs frame. The default is to use `read-file-name' for choosing
-;; a single file and a pair of Dired buffers for choosing multiple files.
-;;
+;; using Emacs.
 ;;
 ;;; Code:
 (require 'compat)
@@ -27,27 +25,24 @@
 
 ;;; Variables
 ;;;; Keymaps
-(defvar filechooser-dired-overriding-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map (kbd "C-c C-c") #'exit-recursive-edit)
-    (define-key map (kbd "C-c C-k") #'filechooser-abort )
-    (define-key map (kbd "C-f") #'filechooser-toggle-filter)
-    (define-key map [remap abort-recursive-edit] #'filechooser-abort)
-    map)
-  "Keymap used as `override-global-map' for Dired based file selection.")
+(defvar-keymap filechooser-dired-overriding-map
+  :doc "Keymap used as `override-global-map' for Dired based file selection."
+  "C-c C-c" #'exit-recursive-edit
+  "C-c C-k" #'filechooser-abort
+  "C-c C-s" #'filechooser-dired-selection-mode
+  "C-c C-u" #'filechooser-dired-clear-selection
+  "C-f" #'filechooser-toggle-filter
+  "<remap> <abort-recursive-edit>" #'filechooser-abort)
 
-(defvar filechooser-mininuffer-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map (kbd "C-f") #'filechooser-toggle-filter)
-    (define-key map [remap abort-recursive-edit] #'filechooser-abort)
-    map))
+(defvar-keymap filechooser-mininuffer-map
+  "C-f" #'filechooser-toggle-filter
+  "<remap> <abort-recursive-edit>" #'filechooser-abort)
 
-(defvar filechooser-multiple-selection-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map (kbd "M-TAB") #'filechooser-multiple-continue)
-    (define-key map (kbd "M-RET") #'filechooser-multiple-finalize-current-selection)
-    (define-key map (kbd "M-a") #'filechooser-multiple-select-all)
-    (make-composed-keymap map filechooser-mininuffer-map)))
+(defvar-keymap filechooser-multiple-selection-map
+  :parent filechooser-mininuffer-map
+  "M-TAB" #'filechooser-multiple-continue
+  "M-RET" #'filechooser-multiple-finalize-current-selection
+  "M-a" #'filechooser-multiple-select-all)
 
 ;;;; Custom variables
 (defgroup filechooser nil
@@ -76,8 +71,8 @@ If it is nil the selected frame is used instead."
   :type 'boolean)
 
 (defcustom filechooser-filters `(("Directories" filechooser-file-directory-p . t)
-                              ("Elisp files" ,(rx ".el" eos))
-                              ("Not dot files" ,(rx bos (not ?.))))
+                                 ("Elisp files" ,(rx ".el" eos))
+                                 ("Not dot files" ,(rx bos (not ?.))))
   "An alist of (NAME FILTER . BOOL).
 NAME should describe the filter which can either be a regexp
 or else a predicate function which takes a filename as argument.
@@ -112,14 +107,24 @@ I.e. the key that binds the equivalent of `exit-minibuffer' for the completion
 UI of choice: usually RET."
   :type (if (> emacs-major-version 27) 'key 'key-sequence))
 
+(defcustom filechooser-dired-marker ?>
+  "The character to mark selected files in `filechooser-with-dired'."
+  :type 'character)
+
+(defcustom filechooser-dired-selection-debounce 0.2
+  "Seconds to wait before updating selection buffer in `filechooser-with-dired'."
+  :type 'float)
+
 ;;;; Others
 (defvar filechooser-current-operation nil
   "When filechooser is active, this variable is set to the command being used.")
 
 ;;;; Internal Variables
 (defvar filechooser--filters nil)
-(defvar filechooser--selection nil)
+(defvar filechooser--active-filters nil)
 (defvar filechooser--multiple-selection nil)
+(defvar filechooser--dired-selection nil)
+(defvar filechooser--dired-buffers nil)
 
 ;;; Filters
 (defun filechooser--filters-group-fn (cand transform)
@@ -137,6 +142,16 @@ See Info node `(elisp) Programmed Completion' for CAND and TRANSFORM."
                         (expand-file-name name (dired-current-directory))
                       name)))
 
+(defun filechooser--filters (filters)
+  "Return FILTERS added to `filechooser-filters'."
+  (cl-delete-duplicates (append filechooser-filters filters)
+                        :test #'equal :key #'car))
+
+(defun filechooser--active-filters ()
+  "Return active filters."
+  (delq nil (mapcar (lambda (flt) (if (cddr flt) (cadr flt)))
+                    filechooser--filters)))
+
 (defun filechooser-toggle-filter (arg)
   "Toggle a filter.
 With prefix ARG toggle multiple filters using `completing-read-multiple'."
@@ -154,11 +169,10 @@ With prefix ARG toggle multiple filters using `completing-read-multiple'."
                             #'table nil t)))))
       (dolist (name names)
         (cl-callf not (cdr (alist-get name filechooser--filters nil nil #'equal))))
+      (setq filechooser--active-filters (filechooser--active-filters))
       (if (minibufferp)
           (throw 'continue t)
-        (dolist (buf (match-buffers
-                      `(derived-mode . dired-mode)
-                      (frame-parameter (selected-frame) 'buffer-list)))
+        (dolist (buf filechooser--dired-buffers)
           (with-current-buffer buf (jit-lock-refontify)))))))
 
 (defun filechooser--make-filters (opts)
@@ -178,18 +192,17 @@ With prefix ARG toggle multiple filters using `completing-read-multiple'."
       (cl-callf not (cdr (alist-get (car current) regex-filters nil nil #'equal))))
     (nreverse regex-filters)))
 
-(defun filechooser--filters-predicate (filters)
-  "Make a predicate out of FILTERS."
-  (lambda (name)
-    (catch 'match
-      (dolist (filter filters)
-        (when (cond
-               ((stringp filter)
-                (string-match filter name))
-               ((functionp filter)
-                (funcall filter name))
-               ((error "Unknown filter %S" filter)))
-          (throw 'match t))))))
+(defun filechooser--filters-predicate (name)
+  "Return non-nil if NAME matches an active filter."
+  (catch 'match
+    (dolist (filter filechooser--active-filters)
+      (when (cond
+             ((stringp filter)
+              (string-match filter name))
+             ((functionp filter)
+              (funcall filter name))
+             ((error "Unknown filter %S" filter)))
+        (throw 'match t)))))
 
 ;;; Utility definitions
 (defmacro filechooser--maybe-with-new-frame (minibuffer &rest body)
@@ -252,8 +265,8 @@ See Info node `(elisp) Programmed Completion' for STR, PRED and ACTION."
                  (group-function . filechooser--multiple-group-function)))
     (_ (completion-file-name-table str pred action))))
 
-(defun filechooser--read-file-name-1 (prompt &optional mustmatch filters dir default)
-  "Read a filename with PROMPT and predicate made from FILTERS.
+(defun filechooser--read-file-name-1 (prompt &optional mustmatch dir default)
+  "Read a filename with PROMPT and predicate made from active filters.
 MUSTMATCH and DIR are as in `read-file-name'.  DEFAULT is the default filename.
 If MULTIPLE is non-nil `completing-read-multiple' is used."
   (catch 'continue
@@ -262,8 +275,7 @@ If MULTIPLE is non-nil `completing-read-multiple' is used."
                                                    (current-local-map)))
           (when dir (setq default-directory dir)))
       (read-file-name
-       prompt dir default mustmatch nil
-       (when filters (filechooser--filters-predicate filters))))))
+       prompt dir default mustmatch nil #'filechooser--filters-predicate))))
 
 (defun filechooser--handle-exisiting-file (filename &optional dir filters)
   "Handle an existing FILENAME according to `filechooser-save-existing-files'.
@@ -290,18 +302,17 @@ are the filters to use in that case."
   "Read a filename with PROMPT and predicate made from FILTERS.
 MUSTMATCH and DIR are as in `read-file-name'.  DEFAULT is the default filename.
 If MULTIPLE is non-nil `completing-read-multiple' is used."
-  (setq filechooser--filters (cl-delete-duplicates
-                              (append filechooser-filters filters)
-                              :test #'equal :key #'car))
-  (let ((result t))
+  (let* ((result t)
+         (filechooser--filters (filechooser--filters filters))
+         (filechooser--active-filters (filechooser--active-filters)))
     (while (eq t result)
       (when (minibufferp nil t)
         (abort-minibuffers))
-      (setq result (filechooser--read-file-name-1
-                    prompt mustmatch
-                    (delq nil (mapcar (lambda (flt) (if (cddr flt) (cadr flt)))
-                                      filechooser--filters))
-                    dir default)))
+      (setq result (filechooser--read-file-name-1 prompt mustmatch dir default)))
+    (when (equal result default)
+      (setq result (expand-file-name default dir)))
+    (when (and default (file-directory-p result))
+      (setq result (expand-file-name default result)))
     (if (or mustmatch (not (file-exists-p result)))
         result
       (filechooser--handle-exisiting-file result dir filters))))
@@ -349,14 +360,10 @@ files which satisfy one of the active filters from FILTERS or
 (defun filechooser--multiple-read-file-name (prompt &optional dir)
   "Read a filename with PROMPT and starting from DIR.
 MAP contains additional key bindigs."
-  (let ((result t)
-        filters)
+  (let ((result t))
     (while (eq t result)
       (when (minibufferp nil t)
         (abort-minibuffers))
-      (setq filters (delq nil (mapcar
-                               (lambda (flt) (if (cddr flt) (cadr flt)))
-                               filechooser--filters)))
       (setq result
             (catch 'continue
               (minibuffer-with-setup-hook
@@ -364,7 +371,7 @@ MAP contains additional key bindigs."
                                         filechooser-multiple-selection-map
                                         (current-local-map))))
                 (completing-read prompt #'filechooser--multiple-loop-table
-                                 (filechooser--filters-predicate filters) t
+                                 #'filechooser--filters-predicate t
                                  (abbreviate-file-name dir)
                                  'file-name-history)))))
     result))
@@ -377,13 +384,11 @@ PROMPT is the minibuffer prompt. DIR is the directory where selection starts.
 FILTERS take the same form as elements of `filechooser-filters'. Only those
 files which satisfy one of the active filters from FILTERS or
 `filechooser-filters' are presented for completions."
-  (setq filechooser--filters
-        (cl-delete-duplicates
-         (append filechooser-filters filters)
-         :test #'equal :key #'car))
   (setq dir (file-name-as-directory
              (expand-file-name (or dir default-directory))))
-  (let (selected filechooser--multiple-selection)
+  (let* ((filechooser--filters (filechooser--filters filters))
+         (filechooser--active-filters (filechooser--active-filters))
+         selected filechooser--multiple-selection)
     (filechooser--maybe-with-new-frame only
       (while (setq dir
                    (catch 'done
@@ -416,107 +421,161 @@ is used, othewise the selected frame is used.  PROMPT and DIR are as in
         names))))
 
 ;;; Dired based selection
-(let (marked unmarked timer)
+(define-minor-mode filechooser-dired-selection-mode
+  "Minor mode that is activated during Dired based file selection.
+It can be disabled to temporarily pause the selection and do other things
+without exiting file selection."
+  :global t
+  :keymap nil
+  (if filechooser-dired-selection-mode
+      (progn
+        (put 'filechooser-dired-marker 'filechooser--original dired-marker-char)
+        (setq dired-marker-char filechooser-dired-marker))
+    (setq dired-marker-char
+          (or (get 'filechooser-dired-marker 'filechooser--original)
+              dired-marker-char))))
+
+(declare-function filechooser--adjust-selection-buffer nil)
+(declare-function filechooser--process-changed-marks nil)
+(declare-function filechooser-dired-clear-selection nil)
+
+(let ((selection (make-hash-table :test #'equal))
+      timer)
+
   (defun filechooser--adjust-selection-buffer ()
-    (when (buffer-live-p (cdr filechooser--selection))
-      (with-current-buffer (cdr filechooser--selection)
-        (cl-callf cl-set-difference (cdr dired-directory) unmarked :test #'equal)
-        (when marked
-          (cl-callf nreverse (cdr dired-directory))
-          (dolist (file marked)
-            (cl-pushnew file (cdr dired-directory) :test #'equal))
-          (cl-callf nreverse (cdr dired-directory)))
+    (when (buffer-live-p (cdr filechooser--dired-selection))
+      (with-current-buffer (cdr filechooser--dired-selection)
+        (setf (cdr dired-directory) (hash-table-keys selection))
         (revert-buffer)))
-    (setq marked nil unmarked nil timer nil))
+    (setq timer nil))
+
+  (defun filechooser-dired-clear-selection (&optional beg end)
+    "Remove files from BEG to END from selection."
+    (interactive)
+    (with-current-buffer (cdr filechooser--dired-selection)
+      (let ((dired-marker-char (get 'filechooser-dired-marker 'filechooser--original))
+            (filechooser-dired-selection-mode nil))
+        (if (eq t end)
+            (dired-mark beg t)
+          (dired-mark-files-in-region (or beg (point-min)) (or end (point-max))))
+        (let ((files (delq nil (dired-map-over-marks
+                                (dired-get-filename nil t) nil))))
+          (dolist (buf filechooser--dired-buffers)
+            (when (buffer-live-p buf)
+              (with-current-buffer buf
+                (save-excursion
+                  (dolist (file files)
+                    (when (dired-goto-file file)
+                      (dired-unmark nil)))))))
+          (dolist (file files)
+            (remhash file selection)))))
+    (filechooser--adjust-selection-buffer))
 
   (defun filechooser--process-changed-marks (beg end _length)
     "Deal with change in mark from BEG to END."
-    (when (and (derived-mode-p 'dired-mode)
-               (eq (1+ beg) end)
-               (not (invisible-p (1- (pos-eol)))))
+    (when (and filechooser-dired-selection-mode
+               (derived-mode-p 'dired-mode)
+               (eq (1+ beg) end))
       (save-excursion
         (goto-char beg)
-        (when (and (re-search-forward dired-re-mark end t)
-                   (eq (preceding-char) dired-marker-char))
-          (push (dired-get-filename nil t) marked))
-        (when (re-search-forward dired-re-maybe-mark (1+ end) t)
-          (push (dired-get-filename nil t) unmarked))
+        (cond
+         ((or (invisible-p (1- (pos-eol)))
+              (not (filechooser--filters-predicate
+                    (dired-get-filename 'no-dir t))))
+          (dired-unmark nil))
+         ((and (re-search-forward dired-re-mark end t)
+               (eq (preceding-char) dired-marker-char))
+          (puthash (dired-get-filename nil t) t selection))
+         ((and (re-search-forward dired-re-maybe-mark (1+ end) t))
+          (remhash (dired-get-filename nil t) selection)))
         (unless timer
           (setq timer (run-with-timer
-                       0.2 nil 'filechooser--adjust-selection-buffer)))))))
+                       filechooser-dired-selection-debounce nil
+                       #'filechooser--adjust-selection-buffer)))))))
+
+(defun filechooser-dired-deselect (arg &optional interactive)
+  "Deselect the file at point in selection buffer.
+See `dired-mark' for ARG and INTERACTIVE."
+  (interactive (list current-prefix-arg t))
+  (filechooser-dired-clear-selection arg interactive))
+
+(defun filechooser--dired-setup-buffer (_)
+  "Setup the current buffer for file selection."
+  (when (and (derived-mode-p 'dired-mode)
+             (not (memq (current-buffer)
+                        `(,(cdr filechooser--dired-selection)
+                          ,@filechooser--dired-buffers))))
+    (add-hook 'jit-lock-functions #'filechooser--dired-jit-filter 95 t)
+    (jit-lock-mode t)
+    (add-to-invisibility-spec 'filechooser-filter)
+    (cl-pushnew (current-buffer) filechooser--dired-buffers)))
 
 (defun filechooser-dired (&optional dir filters)
   "Select some files using Dired.
 Running this command pops a Dired for directory DIR, and enters a recursive
 editing session.  FILTERS are in the format of `filechooser-filters'."
-  (unless (and filechooser--selection
-               (file-directory-p (car filechooser--selection)))
-    (setq filechooser--selection (list (make-temp-file "filechooser-selection-" t))))
-  (let ((overriding-map `((t . ,filechooser-dired-overriding-map)))
-        (apply-filters (lambda (_)
-                         (when (and (derived-mode-p 'dired-mode)
-                                    (not (eq (current-buffer)
-                                             (cdr filechooser--selection)))
-                                    (not (memq 'filechooser--dired-jit-filter
-                                               jit-lock-functions)))
-                           (add-hook 'jit-lock-functions
-                                     #'filechooser--dired-jit-filter 95 t)
-                           (jit-lock-mode t)
-                           (add-to-invisibility-spec 'filechooser-filter))))
-        (selection-buffer (progn (setcdr filechooser--selection nil)
-                                 (dired-noselect filechooser--selection))))
+  (unless (and filechooser--dired-selection
+               (file-directory-p (car filechooser--dired-selection)))
+    (setq filechooser--dired-selection (list (make-temp-file "filechooser-selection-" t))))
+  (let* ((overriding-map `((t . ,filechooser-dired-overriding-map)))
+         (filechooser--filters (filechooser--filters filters))
+         (filechooser--active-filters (filechooser--active-filters))
+         filechooser--dired-buffers)
     (save-window-excursion
       (unwind-protect
-          (progn (setcdr filechooser--selection
-                         (dired-noselect (list (car filechooser--selection))))
-                 (display-buffer selection-buffer '(display-buffer-in-side-window
-                                                    (side . left) (window-width . 0.3)))
-                 (when-let ((win (get-buffer-window (cdr filechooser--selection))))
-                   (select-window win)
-                   (redisplay))
-                 (with-current-buffer (cdr filechooser--selection)
-                   (setq mode-line-format " Selected files")
-                   (dired-hide-details-mode)
-                   (add-hook 'jit-lock-functions #'filechooser--dired-jit-abbreviate 95 t)
-                   (jit-lock-mode t))
-                 (push overriding-map emulation-mode-map-alists)
-                 (add-hook 'window-buffer-change-functions apply-filters)
-                 (add-hook 'after-change-functions 'filechooser--process-changed-marks)
-                 (setq filechooser--filters (append filechooser-filters filters))
-                 (other-window 1)
-                 (dired (or dir default-directory))
-                 (funcall apply-filters nil)
-                 (unless (recursive-edit)
-                   (with-current-buffer (cdr filechooser--selection)
-                     (cdr dired-directory))))
+          (progn
+            (pop-to-buffer
+             (setcdr filechooser--dired-selection
+                     (dired-noselect (list (car filechooser--dired-selection))))
+             '(display-buffer-in-side-window
+               (side . left) (window-width . 0.3)))
+            (filechooser-dired-selection-mode)
+            (filechooser-dired-clear-selection)
+            (redisplay)
+            (with-current-buffer (cdr filechooser--dired-selection)
+              (setq mode-line-format " Selected files")
+              (dired-hide-details-mode)
+              (add-hook 'jit-lock-functions #'filechooser--dired-jit-abbreviate 95 t)
+              (jit-lock-mode t)
+              (use-local-map
+               (define-keymap :parent (current-local-map)
+                 "<remap> <dired-unmark>" #'filechooser-dired-deselect
+                 "<remap> <dired-unmark-all-marks>" #'filechooser-dired-clear-selection)))
+            (push overriding-map emulation-mode-map-alists)
+            (add-hook 'window-buffer-change-functions #'filechooser--dired-setup-buffer)
+            (add-hook 'after-change-functions #'filechooser--process-changed-marks)
+            (other-window 1)
+            (dired (or dir default-directory))
+            (filechooser--dired-setup-buffer nil)
+            (unless (recursive-edit)
+              (filechooser--adjust-selection-buffer)
+              (with-current-buffer (cdr filechooser--dired-selection)
+                (cdr dired-directory))))
         (cl-callf2 delq overriding-map emulation-mode-map-alists)
-        (remove-hook 'window-buffer-change-functions apply-filters)
-        (remove-hook 'after-change-functions 'filechooser--process-changed-marks)
-        (kill-buffer (cdr filechooser--selection))
-        (setcdr filechooser--selection nil)
-        (dolist (buf (match-buffers `(derived-mode . dired-mode) (frame-parameter nil 'buffer-list)))
+        (filechooser-dired-clear-selection)
+        (filechooser-dired-selection-mode -1)
+        (remove-hook 'window-buffer-change-functions #'filechooser--dired-setup-buffer)
+        (remove-hook 'after-change-functions #'filechooser--process-changed-marks)
+        (kill-buffer (cdr filechooser--dired-selection))
+        (setcdr filechooser--dired-selection nil)
+        (dolist (buf filechooser--dired-buffers)
           (with-current-buffer buf
             (jit-lock-unregister #'filechooser--dired-jit-filter)
             (remove-from-invisibility-spec 'filechooser-filter)))))))
 
 (defun filechooser--dired-jit-filter (beg end)
   "Hide files that don't match current filters from BEG to END."
-  (if-let ((active (delq nil (mapcar (lambda (flt) (if (cddr flt) (cadr flt)))
-                                     filechooser--filters))))
-      (progn
-        (setq end (progn (goto-char end) (forward-line 1) (pos-eol)))
-        (setq beg (progn (goto-char beg) (forward-line -1) (goto-char (pos-bol))))
-        (let ((pred (filechooser--filters-predicate active)))
-          (while (< (point) end)
-            (when-let ((name (dired-get-filename 'no-dir t)))
-              (alter-text-property (1- (point)) (pos-eol) 'invisible
-                                   (lambda (val)
-                                     (setq val (ensure-list val))
-                                     (if (funcall pred name)
-                                         (cl-callf2 delq 'filechooser-filter val)
-                                       (cl-pushnew 'filechooser-filter val)))))
-            (forward-line))))
-    (remove-from-invisibility-spec 'filechooser-filter))
+  (setq end (progn (goto-char end) (forward-line 1) (pos-eol)))
+  (setq beg (progn (goto-char beg) (forward-line -1) (goto-char (pos-bol))))
+  (while (< (point) end)
+    (when-let ((name (dired-get-filename 'no-dir t)))
+      (alter-text-property (1- (point)) (pos-eol) 'invisible
+                           (lambda (val)
+                             (setq val (ensure-list val))
+                             (if (filechooser--filters-predicate name)
+                                 (cl-callf2 delq 'filechooser-filter val)
+                               (cl-pushnew 'filechooser-filter val)))))
+    (forward-line))
   `(jit-lock-bounds ,beg . ,end))
 
 (defun filechooser--dired-jit-abbreviate (beg end)
